@@ -1,4 +1,4 @@
-// Copyright (C) 2014-2021 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2023 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -17,12 +17,18 @@
 #include <boost/date_time/posix_time/posix_time.hpp>
 #include <boost/shared_ptr.hpp>
 
+#include <limits>
 #include <utility>
 
 namespace isc {
 
 namespace dhcp {
 
+/// @brief A value used to signal that the interface index was not set.
+/// That means that more than UNSET_IFINDEX interfaces are not supported.
+/// That's fine, since it would have overflowed with UNSET_IFINDEX + 1 anyway.
+constexpr unsigned int UNSET_IFINDEX = std::numeric_limits<unsigned int>::max();
+
 /// @brief RAII object enabling copying options retrieved from the
 /// packet.
 ///
@@ -74,7 +80,7 @@ public:
 
 private:
 
-    /// @brief Holds a pointers to the packets.
+    /// @brief Holds a pair of pointers of the packets.
     std::pair<PktTypePtr, PktTypePtr> pkts_;
 };
 
@@ -171,7 +177,9 @@ public:
     /// @note This buffer is only valid till object that returned it exists.
     ///
     /// @return reference to output buffer
-    isc::util::OutputBuffer& getBuffer() { return (buffer_out_); };
+    isc::util::OutputBuffer& getBuffer() {
+        return (buffer_out_);
+    }
 
     /// @brief Adds an option to this packet.
     ///
@@ -258,12 +266,16 @@ public:
     /// @brief Sets transaction-id value.
     ///
     /// @param transid transaction-id to be set.
-    void setTransid(uint32_t transid) { transid_ = transid; }
+    void setTransid(uint32_t transid) {
+        transid_ = transid;
+    }
 
     /// @brief Returns value of transaction-id field.
     ///
     /// @return transaction-id
-    uint32_t getTransid() const { return (transid_); };
+    uint32_t getTransid() const {
+        return (transid_);
+    }
 
     /// @brief Checks whether a client belongs to a given class.
     ///
@@ -271,25 +283,28 @@ public:
     /// @return true if belongs
     bool inClass(const isc::dhcp::ClientClass& client_class);
 
-    /// @brief Adds packet to a specified class.
+    /// @brief Adds a specified class to the packet.
     ///
-    /// A packet can be added to the same class repeatedly. Any additional
-    /// attempts to add to a class the packet already belongs to, will be
+    /// A class can be added to the same packet repeatedly. Any additional
+    /// attempts to add to a packet the class already added, will be
     /// ignored silently.
     ///
-    /// @note It is a matter of naming convention. Conceptually, the server
-    /// processes a stream of packets, with some packets belonging to given
-    /// classes. From that perspective, this method adds a packet to specified
-    /// class. Implementation wise, it looks the opposite - the class name
-    /// is added to the packet. Perhaps the most appropriate name for this
-    /// method would be associateWithClass()? But that seems overly long,
-    /// so I decided to stick with addClass().
-    ///
     /// @param client_class name of the class to be added
     /// @param required the class is marked for required evaluation
     void addClass(const isc::dhcp::ClientClass& client_class,
                   bool required = false);
 
+    /// @brief Adds a specified subclass to the packet.
+    ///
+    /// A subclass can be added to the same packet repeatedly. Any additional
+    /// attempts to add to a packet the subclass already added, will be
+    /// ignored silently.
+    ///
+    /// @param class_def name of the class definition to be added
+    /// @param subclass name of the subclass to be added
+    void addSubClass(const isc::dhcp::ClientClass& class_def,
+                     const isc::dhcp::ClientClass& subclass);
+
     /// @brief Returns the class set
     ///
     /// @note This should be used only to iterate over the class set.
@@ -301,6 +316,19 @@ public:
         return (!required ? classes_ : required_classes_);
     }
 
+    /// @brief Returns the class set including template classes associated with
+    /// subclasses
+    ///
+    /// @note This should be used only to iterate over the class set.
+    /// @note SubClasses are always last.
+    /// @param required return classes or required to be evaluated classes.
+    /// @return if required is false (the default) the classes the
+    /// packet belongs to else the classes which are required to be
+    /// evaluated.
+    const SubClassRelationContainer& getSubClassesRelations() const {
+        return (subclasses_);
+    }
+
     /// @brief Unparsed data (in received packets).
     ///
     /// @warning This public member is accessed by derived
@@ -326,13 +354,31 @@ protected:
     /// if such option is not present.
     OptionPtr getNonCopiedOption(const uint16_t type) const;
 
+    /// @brief Returns all option instances of specified type without
+    /// copying.
+    ///
+    /// This is a variant of @ref getOptions method, which returns a collection
+    /// of options without copying them. This method should be only used by
+    /// the @ref Pkt6 class and derived classes. Any external callers should
+    /// use @ref getOptions which copies option instances before returning them
+    /// when the @ref Pkt::copy_retrieved_options_ flag is set to true.
+    ///
+    /// @param opt_type Option code.
+    ///
+    /// @return Collection of options found.
+    OptionCollection getNonCopiedOptions(const uint16_t opt_type) const;
+
 public:
 
+    /// @brief Clones all options so that they can be safely modified.
+    ///
+    /// @return A container with option clones.
+    OptionCollection cloneOptions();
+
     /// @brief Returns the first option of specified type.
     ///
     /// Returns the first option of specified type. Note that in DHCPv6 several
     /// instances of the same option are allowed (and frequently used).
-    /// Also see @ref Pkt6::getOptions().
     ///
     /// The options will be only returned after unpack() is called.
     ///
@@ -341,6 +387,15 @@ public:
     /// @return pointer to found option (or NULL)
     OptionPtr getOption(const uint16_t type);
 
+    /// @brief Returns all instances of specified type.
+    ///
+    /// Returns all instances of options of the specified type. DHCPv6 protocol
+    /// allows (and uses frequently) multiple instances.
+    ///
+    /// @param type option type we are looking for
+    /// @return instance of option collection with requested options
+    isc::dhcp::OptionCollection getOptions(const uint16_t type);
+
     /// @brief Controls whether the option retrieved by the @ref Pkt::getOption
     /// should be copied before being returned.
     ///
@@ -357,7 +412,7 @@ public:
     /// Kea configuration. To prevent this, option copying should be
     /// enabled prior to passing the pointer to a packet to a hook library.
     ///
-    /// Note that only only does this method causes the server to copy
+    /// Not only does this method cause the server to copy
     /// an option, but the copied option also replaces the original
     /// option within the packet. The option can be then freely modified
     /// and the modifications will only affect the instance of this
@@ -480,13 +535,13 @@ public:
     /// @brief Sets interface index.
     ///
     /// @param ifindex specifies interface index.
-    void setIndex(int ifindex) {
+    void setIndex(const unsigned int ifindex) {
         ifindex_ = ifindex;
-    };
+    }
 
     /// @brief Resets interface index to negative value.
     void resetIndex() {
-        ifindex_ = -1;
+        ifindex_ = UNSET_IFINDEX;
     }
 
     /// @brief Returns interface index.
@@ -494,13 +549,13 @@ public:
     /// @return interface index
     int getIndex() const {
         return (ifindex_);
-    };
+    }
 
     /// @brief Checks if interface index has been set.
     ///
     /// @return true if interface index set, false otherwise.
     bool indexSet() const {
-        return (ifindex_ >= 0);
+        return (ifindex_ != UNSET_IFINDEX);
     }
 
     /// @brief Returns interface name.
@@ -509,15 +564,19 @@ public:
     /// going to be transmitted.
     ///
     /// @return interface name
-    std::string getIface() const { return (iface_); };
+    std::string getIface() const {
+        return (iface_);
+    }
 
     /// @brief Sets interface name.
     ///
     /// Sets interface name over which packet was received or is
     /// going to be transmitted.
     ///
-    /// @return interface name
-    void setIface(const std::string& iface ) { iface_ = iface; };
+    /// @param iface The interface name
+    void setIface(const std::string& iface) {
+        iface_ = iface;
+    }
 
     /// @brief Sets remote hardware address.
     ///
@@ -592,7 +651,7 @@ public:
     /// This field is public, so the code outside of Pkt4 or Pkt6 class can
     /// iterate over existing classes. Having it public also solves the problem
     /// of returned reference lifetime. It is preferred to use @ref inClass and
-    /// @ref addClass should be used to operate on this field.
+    /// @ref addClass to operate on this field.
     ClientClasses classes_;
 
     /// @brief Classes which are required to be evaluated.
@@ -603,6 +662,14 @@ public:
     /// and if evaluation status is true added to the previous collection.
     ClientClasses required_classes_;
 
+    /// @brief SubClasses this packet belongs to.
+    ///
+    /// This field is public, so the code outside of Pkt4 or Pkt6 class can
+    /// iterate over existing classes. Having it public also solves the problem
+    /// of returned reference lifetime. It is preferred to use @ref inClass and
+    /// @ref addSubClass to operate on this field.
+    SubClassRelationContainer subclasses_;
+
     /// @brief Collection of options present in this message.
     ///
     /// @warning This public member is accessed by derived
@@ -733,7 +800,7 @@ protected:
     /// Each network interface has assigned an unique ifindex.
     /// It is a functional equivalent of a name, but sometimes more useful, e.g.
     /// when using odd systems that allow spaces in interface names.
-    int64_t ifindex_;
+    unsigned int ifindex_;
 
     /// @brief Local IP (v4 or v6) address.
     ///
@@ -796,7 +863,92 @@ private:
 /// @brief A pointer to either Pkt4 or Pkt6 packet
 typedef boost::shared_ptr<isc::dhcp::Pkt> PktPtr;
 
-}; // namespace isc::dhcp
-}; // namespace isc
+/// @brief RAII object enabling duplication of the stored options and restoring
+/// the original options on destructor.
+///
+/// This object enables duplication of the stored options and restoring the
+/// original options on destructor. When the object goes out of scope, the
+/// initial options are restored. This is applicable in cases when the server is
+/// going to invoke a callout (hook library) where the list of options in the
+/// packet will be modified. This can also be used to restore the initial
+/// suboptions of an option when the suboptions are changed (e.g. when splitting
+/// long options and suboptions). The use of RAII object eliminates the need for
+/// explicitly copying and restoring the list of options and is safer in case of
+/// exceptions thrown by callouts and a presence of multiple exit points.
+class ScopedSubOptionsCopy {
+public:
+
+    /// @brief Constructor.
+    ///
+    /// Creates a copy of the initial options on an option.
+    ///
+    /// @param opt Pointer to the option.
+    ScopedSubOptionsCopy(const OptionPtr& opt) : option_(opt) {
+        if (opt) {
+            options_ = opt->getMutableOptions();
+        }
+    }
+
+    /// @brief Destructor.
+    ///
+    /// Restores the initial options on a packet.
+    ~ScopedSubOptionsCopy() {
+        if (option_) {
+            option_->getMutableOptions() = options_;
+        }
+    }
+
+private:
+
+    /// @brief Holds a pointer to the option.
+    OptionPtr option_;
+
+    /// @brief Holds the initial options.
+    OptionCollection options_;
+};
+
+/// @brief RAII object enabling duplication of the stored options and restoring
+/// the original options on destructor.
+///
+/// This object enables duplication of the stored options and restoring the
+/// original options on destructor. When the object goes out of scope, the
+/// initial options are restored. This is applicable in cases when the server is
+/// going to invoke a callout (hook library) where the list of options in the
+/// packet will be modified. The use of RAII object eliminates the need for
+/// explicitly copying and restoring the list of options and is safer in case of
+/// exceptions thrown by callouts and a presence of multiple exit points.
+///
+/// @tparam PktType Type of the packet, e.g. Pkt4, Pkt6, Pkt4o6.
+template<typename PktType>
+class ScopedPktOptionsCopy {
+public:
+
+    /// @brief Constructor.
+    ///
+    /// Creates a copy of the initial options on a packet.
+    ///
+    /// @param pkt Pointer to the packet.
+    ScopedPktOptionsCopy(PktType& pkt) : pkt_(pkt), options_(pkt.options_) {
+        pkt_.options_ = pkt_.cloneOptions();
+    }
+
+    /// @brief Destructor.
+    ///
+    /// Restores the initial options on a packet.
+    ~ScopedPktOptionsCopy() {
+        pkt_.options_ = options_;
+    }
+
+private:
+
+    /// @brief Holds a reference to the packet.
+    PktType& pkt_;
+
+    /// @brief Holds the initial options.
+    OptionCollection options_;
+};
+
+} // end of namespace isc::dhcp
+} // end of namespace isc
 
 #endif
