@@ -22,7 +22,7 @@ prefix="/usr/local"
 if test -f "${prefix}/share/kea/scripts/admin-utils.sh"; then
     . "${prefix}/share/kea/scripts/admin-utils.sh"
 else
-    . "/pobj/kea-2.4.1-postgresql/kea-2.4.1/src/bin/admin/admin-utils.sh"
+    . "/build/tmp/pobj/kea-2.4.2-postgresql/kea-2.4.2/src/bin/admin/admin-utils.sh"
 fi
 
 VERSION=$(pgsql_version "$@")
